import java.util.*;

public class AlphaBeta {
	private long timeLimit;
	private int SELF_PIECE = -15;
	private int SELF_KING = -35;
	private int OPP_PIECE = 15;
	private int OPP_KING = 35;
	
	private int currentPlayer = 0;
	private String theBestMove;

	private int MAX_DEPTH = 10;

	//don't want AI player to be in the king row, so bad value
	private int WHITESELF_SQUARES[] = 
	{0,-5,-5,-5,-5, //Back Row
		-1,-1,-1,-1,       //Row 2
		-2,-3,-3,-2,       //Row 3
		-3,-4,-4,-3,       //Row 4
		-4,-5,-5,-4,      //Row 5
		-5,-5,-5,-5,       //Row 6
		-7,-6,-6,-7,       //Row 7
		-15,-15,-15,-15};  //King Row

	//however, we want the human to be in the AI player's back row
	//so high value
	private int WHITEOPP_SQUARES[] = 
	{0,15,15,15,15,  //Back Row
		7, 6, 6, 7,     //Row 2
		6, 5, 5, 6,     //Row 3
		4, 5, 5, 4,     //Row 4
		3, 4, 4, 3,     //Row 5
		2, 3, 3, 2,     //Row 6
		1, 1, 1, 1,     //Row 7
		5, 5, 5, 5};    //King Row

	private int REDOPP_SQUARES[] = 
	{0, 5, 5, 5, 5, //Back Row
		1, 1, 1, 1,       //Row 2
		2, 3, 3, 2,       //Row 3
		3, 4, 4, 3,       //Row 4
		4, 5, 5, 4,      //Row 5
		5, 5, 5, 5,       //Row 6
		7, 6, 6, 7,       //Row 7
		15, 15, 15, 15};  //King Row

	//however, we want the human to be in the AI player's back row
	//so high value
	private int REDSELF_SQUARES[] = 
	{0,-15,-15,-15,-15,  //Back Row
		-7, -6, -6, -7,     //Row 2
		-6, -5, -5, -6,     //Row 3
		-4, -5, -5, -4,     //Row 4
		-3, -4, -4, -3,     //Row 5
		-2, -3, -3, -2,     //Row 6
		-1, -1, -1, -1,     //Row 7
		-5, -5, -5, -5};    //King Row

	public AlphaBeta()
	{
		currentPlayer = Checkers.getCurrentPlayer() - 1;
		timeLimit = 60000;
	}

	public AlphaBeta(int time)
	{
		currentPlayer = Checkers.getCurrentPlayer() - 1;
		timeLimit = time;
	}

	public String bestMove (Board currentBoard) {
		AlphaBetaSearch (currentBoard);
		return theBestMove;
	}

	public void AlphaBetaSearch(Board currentBoard)
	{
		Board tempBoard = new Board (currentBoard);
		
		Vector jumps = tempBoard.forcedJumps (Checkers.getCurrentPlayer());
		
		if (jumps.size() == 0) {	
			int maxBoardValue = MaxValue(tempBoard,Integer.MIN_VALUE,Integer.MAX_VALUE, 0);
		} else {
			TreeSet js = new TreeSet();
			for (int i=0; i < jumps.size(); i++) {
				String[] jump = ((String) jumps.elementAt(i)).split("-");

				int place = new Integer(jump[0]).intValue();
				int endPlace = new Integer(jump[1]).intValue();
				Piece p = tempBoard.pieceAt(place);
				JumpSuccessor jumpSucc = new JumpSuccessor (p, (String) jumps.elementAt(i), 0);

				Board tBoard = new Board (tempBoard);
				for (int j = 0; j < 4; j++) {
					Piece t = new Piece (endPlace, Checkers.getCurrentPlayer());
					if (p.isKing()) {
						t.kingMe();
					}
					tBoard.addPiece (t);
					Vector nextJumps = tBoard.jumpsForPiece (t, Checkers.getCurrentPlayer());
				
					if (nextJumps.size() == 0) {
						j = 15;
					} else {
						String nextJump = (String) nextJumps.elementAt(i);
						jumpSucc.addJump (nextJump);
						
						
						String[] nJ = nextJump.split ("-");
						Piece jumped = tBoard.jumpedPiece (new Integer(nJ[0]).intValue(),new Integer(nJ[1]).intValue());
						tBoard.removePiece (jumped);
						endPlace = new Integer(nJ[1]).intValue();
					}
					tBoard.removePiece (t);
				}
				
				js.add (jumpSucc);
			}
			
			theBestMove = ((JumpSuccessor) js.first()).toString();
		}
	}
    
	public Board getMove(int maxBoardValue)
	{
		//again need method to update the board
		Board newBoard = null; //= new board;
		return newBoard;    
	}

	public boolean terminalTest(Board currentBoard, int depth, int player)
	{
		/*we may have to alter this b/c we are not going to reach a
		* state where there are no pieces left in the first couple of searches
		*/
		if (depth >= MAX_DEPTH) {
			return true;
		}

		//number of computer pieces
		int numPieces = currentBoard.numPieces(0);   //0 as opposed to 1
		int numHumanPieces = currentBoard.numPieces(1);//human pieces
		if (numPieces == 0 || numHumanPieces == 0)
			return true;
		else if (!currentBoard.legalMovesLeft(player))   //need to check for any moves left
			return true;
		else
			return false;
	}
    
	public int h(String move, Board currentBoard)
	{
		/*simple heuristic
		int computerPieces = currentBoard.numPieces(0);
		int humanPieces = currentBoard.numPieces(1);
		return humanPieces - computerPieces;
		*/
       
		//second heuristic
		int hVal = 0;
		String[] placesStr = move.split ("-");
		int pieceLocation = new Integer(placesStr[placesStr.length-1]).intValue();

		Piece destination = currentBoard.pieceAt (pieceLocation);
		Piece start = currentBoard.pieceAt (new Integer (placesStr[0]).intValue());
		
		int myPlayerNumber = Checkers.currentMover().getPlayerNumber();
		
		if(destination == null) {
			//I'm on top of the board.
			if (myPlayerNumber == 0) {
				if(!start.ownedBy(currentPlayer)) {
					if(start.isKing()) {
						hVal += SELF_KING;
					} else {
						hVal = SELF_PIECE + WHITESELF_SQUARES[pieceLocation];
					}
				} else {
					if(start.isKing()) {
						hVal += OPP_KING;
					} else {
						hVal = OPP_PIECE + WHITEOPP_SQUARES[pieceLocation];
					}
				}
			} else {
				if(!start.ownedBy(currentPlayer)) {
					if(start.isKing()) {
						hVal += SELF_KING;
					} else {
						hVal = OPP_PIECE + REDSELF_SQUARES[pieceLocation];
					}
				} else {
					if(start.isKing()) {
						hVal += OPP_KING;
					} else {
						hVal = OPP_PIECE + REDOPP_SQUARES[pieceLocation];
					}
				}
			}
		}
		return hVal;
	}
    
	public int MaxValue(Board boardState, int alpha, int beta, int depth)
	{
		currentPlayer = (currentPlayer + 1) % 2;		
		if(terminalTest(boardState, depth, currentPlayer))
		{
			return 100;
			//return h(boardState);  //has to take string too??!?
		}

		int utility = Integer.MIN_VALUE;
		Vector successor = makeSuccessors(boardState, currentPlayer);
		if (successor.size() == 0) {
			return 100;
		}
		MoveSuccessor nextMove = (MoveSuccessor) successor.lastElement();
		String move = nextMove.getMove();

		Board succBoard = new Board(boardState);
		succBoard.movePiece (nextMove.getPiece(), move);
		utility = Max(utility, MinValue(succBoard,alpha,beta, depth+1));
		if(utility >= beta)
			return utility;
		alpha = Max(alpha, utility);
		theBestMove = move;
		return utility;
	}
    
	public int MinValue(Board boardState, int alpha, int beta, int depth)
	{  
		currentPlayer = (currentPlayer + 1) % 2;
		if (terminalTest(boardState, depth, currentPlayer))
		{
			return -100;
			//return h(boardState); //has to take string too??!?
		}

		int utility = Integer.MAX_VALUE;
		Vector successor = makeSuccessors(boardState, currentPlayer);
		if (successor.size() == 0) {
			return -100;
		}
		MoveSuccessor nextMove = (MoveSuccessor) successor.firstElement();
		String move = nextMove.getMove();

		Board succBoard = new Board (boardState); //(Board) ;
		succBoard.movePiece (nextMove.getPiece(), move);
		utility = Min(utility, MaxValue(succBoard,alpha,beta, depth+1));
		if (utility <= alpha)
			return utility;
		beta = Min(beta, utility);
		return utility;
	}
    
	public int Max(int num1, int num2)
	{
		if (num1>=num2)
			return num1;
		else
			return num2;
	}

	public int Min(int num1, int num2)
	{
		if (num1 <= num2)
			return num1;
		else 
			return num2;
	}
    
	public Vector makeSuccessors (Board state, int playerNum) {
		Vector successors = new Vector();
		
		Vector myPieces = state.playersPieces (playerNum);
		for (int i=0; i < myPieces.size(); i++) {
			Piece p = (Piece) myPieces.elementAt(i);
			Vector moves = state.movesForPiece (p, playerNum);
			
			for (int k=0; k < moves.size(); k++) {
				int moveVal = h ((String) moves.elementAt(k), state);
				successors.add (new MoveSuccessor (p, (String) moves.elementAt(k), moveVal));
			}
		}
		
		Object[] stuff = successors.toArray();
		Arrays.sort (stuff);

		Vector newVec = new Vector();		
		
		for (int i=0; i < stuff.length; i++) {
			newVec.add (stuff[i]);			
		}
		return newVec;
	}
		
   
	public TreeSet pieceSuccessors(Piece piece, Board currentBoard)
	{
		TreeSet successorsForPiece = new TreeSet();
		String nextMove;
		int moveValue;
		MoveSuccessor pieceSucc;
		Vector move = new Vector();
		if(piece.ownedBy(0)) {
			move = currentBoard.movesForPiece(piece,0);  //playernum? AI
		} else {
			move = currentBoard.movesForPiece(piece,1);  //Human?
		}

		for (int i = 0; i < move.size(); i++) {
			nextMove = (String) move.get(i);    
			moveValue = h(nextMove,currentBoard);
			pieceSucc = new MoveSuccessor(piece,nextMove,moveValue);
			successorsForPiece.add(pieceSucc);
		}
		return successorsForPiece;
	}
}
